"""
The ``Date`` class.
"""

import datetime as _dt
import pytz

from .misc import classproperty
from .tz import get_timezone


################################################################################
# Constants

_dtdate = _dt.date

_FORMATS = ['%Y%m%d', '%Y-%m-%d']

_DATE_ALIASES = {
    'MIN': _dtdate.min,
    'MAX': _dtdate.max,
}

_DELTA_MAP = {
    'yest': -1,
    'yesterday': -1,
    'today': 0,
    'tomorrow': +1,
}


################################################################################
# The Date class

class Date(_dtdate):
    """
    A date, in yyyymmdd format.

    This is a subclass of ``datetime.date``, and is mostly (but not fully) compatible with it.

    This class supports flexible creation of ``Date`` objects, e.g.: from a yyyymmdd string, from
    delta (number of days relative to today), aliases ('yesterday', 'today', 'MIN', 'MAX', etc.),
    from a ``datetime.date`` object, etc.

    Convenient date arithmetic: ``Date +/- int => Date``, ``Date - Date => int``.

    """

    FORMAT = _FORMATS[0]

    # ===============================================================================================
    # ctor
    # ===============================================================================================

    def __new__(cls, *args, **kwargs):
        if len(args) == 1 and not kwargs:
            x = args[0]
            d = _to_dt_date(x)
            if d is not None:
                return cls.fromordinal(d.toordinal())
            else:
                raise ValueError('invalid date: %r' % x)
        return super().__new__(cls, *args, **kwargs)

    # ===============================================================================================
    #  adjusted datetime.date interface
    # ===============================================================================================

    def replace(self, *args, **kwargs):
        return type(self)(super().replace(*args, **kwargs))

    @classproperty
    def min(cls):
        """ The earliest date supported. """
        return cls(super().min)

    @classproperty
    def max(cls):
        """ The latest date supported. """
        return cls(super().max)

    @classmethod
    def today(cls):
        """ Today's date, in current timezone. """
        return cls(_today())

    @classmethod
    def utctoday(cls):
        """ Today's date, in UTC. """
        return cls(_utctoday())

    def __str__(self):
        return self.strftime(self.FORMAT)

    def __repr__(self):
        return '%s(%s)' % (type(self).__name__, self)

    # ===============================================================================================
    # date arithmetic
    # ===============================================================================================

    def __add__(self, days):
        """
        >>> Date(20120130) + 2
        Date(20120201)
        """
        return self.fromordinal(self.toordinal() + days)

    def __sub__(self, x):
        """
        >>> Date(20120130) - 2
        Date(20120128)
        >>> Date(20120130) - Date(20120128)
        2
        """
        if isinstance(x, int):
            # subtract days
            return self + (-x)
        elif isinstance(x, Date):
            # difference between dates, in days
            return (self.as_datetime_date() - x.as_datetime_date()).days
        else:
            raise TypeError(x)

    # ===============================================================================================
    # misc
    # ===============================================================================================

    def year_start(self):
        """
        >>> Date(20120830).year_start()
        Date(20120101)
        """
        return self.replace(month=1, day=1)

    def year_end(self):
        """
        >>> Date(20120830).year_end()
        Date(20121231)
        """
        return self.replace(year=self.year + 1).year_start() - 1

    def month_start(self):
        """
        >>> Date(20120830).month_start()
        Date(20120801)
        """
        return self.replace(day=1)

    def month_end(self):
        """
        >>> Date(20120115).month_end()
        Date(20120131)
        """
        d = self.replace(day=28)
        while d.month == (d + 1).month:
            d += 1
        return d

    def min_timestamp(self, **kwargs):
        """
        First timestamp of the day.
        """
        return _dt.datetime.fromordinal(self.toordinal()).replace(**kwargs)

    def max_timestamp(self, **kwargs):
        """
        Last timestamp of the day.
        """
        return (self + 1).min_timestamp(**kwargs) - _dt.datetime.resolution

    @property
    def yyyy(self):
        """
        A string represeneting the year part.

        >>> Date(19950607).yyyy
        '1995'
        """
        return self.strftime('%Y')

    @property
    def mm(self):
        """
        2-char string represeneting the month part.

        >>> Date(19950607).mm
        '06'
        """
        return self.strftime('%m')

    @property
    def dd(self):
        """
        2-char string represeneting the day part.

        >>> Date(19950607).dd
        '07'
        """
        return self.strftime('%d')

    def as_datetime_date(self):
        """ Convert self to a ``datetime.date`` object. """
        return _dtdate.fromordinal(self.toordinal())


################################################################################
# Private functions

def _today():
    tz = get_timezone()
    if tz is None:
        # local
        return _dtdate.today()
    else:
        # convert utc to TIMEZONE
        t_utc = pytz.utc.localize(_dt.datetime.utcnow())
        t_local = t_utc.astimezone(tz)
        return t_local.date()


def _utctoday():
    return _dt.datetime.utcnow().date()


def _to_dt_date(x):
    if isinstance(x, (_dtdate, Date)):
        return x
    elif isinstance(x, str):
        return _from_string(x)
    elif isinstance(x, bytes):
        # for pickle support
        return _dtdate(x)
    elif isinstance(x, int):
        return _from_int(x)
    else:
        return None


def _from_string(x):
    if not x:
        raise ValueError(x)
    # relative
    delta = _delta_from_string(x)
    if delta is not None:
        return _from_delta(delta)
    # date aliases
    try:
        return _DATE_ALIASES[x]
    except KeyError:
        pass

    # absolute date
    # we zero-pad the year, otherwise dates before year 1000 can parse incorrectly
    x2 = _pad_date_string(x)
    for fmt in _FORMATS:
        try:
            return _dt.datetime.strptime(x2, fmt).date()
        except ValueError:
            continue
    return None


def _delta_from_string(x):
    try:
        return _DELTA_MAP[x.lower()]
    except KeyError:
        pass
    if (x and x[0] in '+-') or x in ['0']:
        try:
            return int(x)
        except (TypeError, ValueError):
            pass
    return None


def _from_int(x):
    if x < 1:
        # negative delta, relative to today, e.g. -1, -500
        return _from_delta(x)
    else:
        # a date as int, e.g. 20120305
        return _dtdate(x // 10000, (x // 100) % 100, x % 100)


def _from_delta(x):
    return _today() + _dt.timedelta(days=x)


def _pad_date_string(x):
    if '-' in x:
        # A yyyy-mm-dd string
        unpadded, _, _ = x.partition('-')
        n = 4
    else:
        # A yyyymmdd string
        unpadded = x
        n = 8
    padlen = max(0, n - len(unpadded))
    return (padlen * '0') + x


################################################################################



"""
The ``DateRange`` class.
"""

from .date import Date
from .misc import classproperty, egcd as _egcd


################################################################################
# The DateRange class

class DateRange(object):
    """
    A `range <https://docs.python.org/3.8/library/functions.html#func-range>`_ -like type, whose
    elements are ``Date``s.

    This class mostly follows the semantics of the builtin ``range`` type.  E.g.,
    stop is exclusive, behavior of steps, negative steps, slicing,
    ``range1 == range2`` iff ``tuple(range1) == tuple(range2)``, etc.

    Creation of ``DateRange`` objects is flexible.  ``start`` and ``stop`` parameters are converted
    to ``Date`` automatically.  See ``Date`` class for values which can be converted.
    """

    DATE_TYPE = Date
    DELIM = ':'
    OTHER_SIDE_MARKER = '%'
    EMPTY_RANGE_STRING = 'NONE'
    FULL_RANGE_STRING = 'ALL'

    # ===============================================================================================
    # ctor
    # ===============================================================================================

    def __init__(self, start, stop, step=1):
        """
        :param start, stop: range's boundries. Automatically converted to Date objects.
        """
        start = self.DATE_TYPE(start)
        stop = self.DATE_TYPE(stop)
        self._ordrange = range(
            self._d2o(start),
            self._d2o(stop),
            step,
        )

    @classmethod
    def empty(cls):
        """ Creates an empty range, """
        return cls(cls.max_date, cls.max_date)

    @classmethod
    def full(cls, step=1):
        """ Creates a range containing all supported dates, """
        dates = [cls.min_date, cls.max_date]
        if step < 0:
            dates = reversed(dates)
        return cls(*dates, step=step)

    @classmethod
    def _from_ordrange(cls, ordrange):
        r = cls(cls.DATE_TYPE.min, cls.DATE_TYPE.min)  # create a stub
        r._ordrange = ordrange  # replace internals
        return r

    # ===============================================================================================
    # range interface
    # ===============================================================================================

    @property
    def start(self):
        """ Same as ``range.start``. """
        return self._o2d(self._ordrange.start)

    @property
    def stop(self):
        """ Same as ``range.stop``. """
        return self._o2d(self._ordrange.stop)

    @property
    def step(self):
        """ Same as ``range.step``. """
        return self._ordrange.step

    def __iter__(self):
        """ Iterate over the ``Date``s in the range. """
        return (self._o2d(o) for o in self._ordrange)

    def __reversed__(self):
        return (self._o2d(o) for o in reversed(self._ordrange))

    def __len__(self):
        return len(self._ordrange)

    def __bool__(self):
        return bool(self._ordrange)

    def __contains__(self, date):
        """
        Is the date contained in the range?
        :param date:
            a Date, a datetime.date, or a datetime.datetime.
            For the latter, it is contained if its date is contained.
        """
        return self._d2o(date) in self._ordrange

    def index(self, date):
        """ Same as ``range.index``. """
        ordinal = self._d2o(date)
        try:
            return self._ordrange.index(ordinal)
        except ValueError:
            raise ValueError('%r is not in range' % date) from None

    def count(self, date):
        """ Same as ``range.count``. """
        return self._ordrange.count(self._d2o(date))

    def __getitem__(self, idx):
        """ Same as ``range.__getitem__``. """
        if isinstance(idx, slice):
            return self._from_ordrange(self._ordrange[idx])
        else:
            return self._o2d(self._ordrange[idx])

    def __hash__(self):
        """ Same as ``range.__hash__``. """
        return hash(self._ordrange)

    def __eq__(self, other):
        try:
            other_ordrange = other._ordrange
        except AttributeError:
            return False
        else:
            return self._ordrange == other_ordrange

    # ===========================================================================
    # str and repr
    # ===========================================================================

    def to_string(self,
                  opening_bracket='[', closing_bracket=')', *,
                  date_format=None, delim=None, step_delim=None, include_step=None,
                  empty_range_string=EMPTY_RANGE_STRING):

        if empty_range_string is not None and not self:
            return empty_range_string

        if delim is None:
            delim = self.DELIM

        if date_format is None:
            d1 = str(self.start)
            d2 = str(self.stop)
        else:
            d1 = self.start.strftime(date_format)
            d2 = self.stop.strftime(date_format)

        tokens = [opening_bracket, d1, delim, d2]

        if (include_step is True) or (include_step is None and self.step != 1):
            if step_delim is None:
                step_delim = delim
            tokens.extend([step_delim, str(self.step)])

        tokens.append(closing_bracket)
        return ''.join(tokens)

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        tokens = [self.start, self.stop]
        if self.step != 1:
            tokens.append(self.step)
        return '%s(%s)' % (type(self).__name__, ', '.join([str(x) for x in tokens]))

    # ===========================================================================
    # from string
    # ===========================================================================

    @classmethod
    def from_string(cls, s, allow_non_range=True):
        """
        Create a DateRange from a string representation.

        >>> DateRange.from_string('20120202:20120303')
        DateRange(20120202, 20120303)
        >>> DateRange.from_string('[20120202:20120303)')
        DateRange(20120202, 20120303)
        >>> DateRange.from_string('20120202:20120303:7')
        DateRange(20120202, 20120303, 7)

        > DateRange.from_string('yest:today')  # if running on 20120705
        DateRange(20120704, 20120705)
        > DateRange.from_string('-1:+1')  # if running on 20120705
        DateRange(20120704, 20120706)
        """

        if s in (cls.FULL_RANGE_STRING, 'FULL'):
            return cls.full()
        if s in (cls.EMPTY_RANGE_STRING, 'EMPTY'):
            return cls.empty()

        # remove brackets:
        if s.startswith('[') and s.endswith(')'):
            s = s[1:-1]

        # split to parts:
        parts = s.split(cls.DELIM)
        n = len(parts)
        if n == 1:
            # a single-date range
            d = cls.DATE_TYPE(s)
            return cls(d, d + 1)

        elif 2 <= n <= 3:
            # start and stop, and possibly step

            def has_marker(p, m=cls.OTHER_SIDE_MARKER):
                return (p == m) or (p.startswith(m) and p[len(m)] in '+-')

            def relative_to(part, rel_to, m=cls.OTHER_SIDE_MARKER):
                offset_str = part[len(cls.OTHER_SIDE_MARKER):]
                if offset_str:
                    # e.g. '%+10'
                    offset = int(offset_str)
                else:
                    # just '%'
                    offset = 0
                return rel_to + offset

            marker1 = has_marker(parts[0])
            marker2 = has_marker(parts[1])
            if marker1:
                d2 = cls.DATE_TYPE(parts[1])
                d1 = relative_to(parts[0], d2)
            elif marker2:
                d1 = cls.DATE_TYPE(parts[0])
                d2 = relative_to(parts[1], d1)
            else:
                d1 = cls.DATE_TYPE(parts[0])
                d2 = cls.DATE_TYPE(parts[1])
            if n == 3:
                step = int(parts[2])
            else:
                step = 1
            return cls(d1, d2, step)

        else:
            raise ValueError('Invalid %s string: %r' % (cls.__name__, s))

    # ===========================================================================
    # misc
    # ===========================================================================

    def replace(self, *, start=None, stop=None, step=None):
        """ Create a range like self, with different parts (start/stop/step) """
        kwargs = {}
        kwargs['start'] = start if start is not None else self.start
        kwargs['stop'] = stop if stop is not None else self.stop
        kwargs['step'] = step if step is not None else self.step
        return type(self)(**kwargs)

    @classproperty
    def min_date(cls):
        """ The earliest date supported. """
        return cls.DATE_TYPE.min

    @classproperty
    def max_date(cls):
        """ The latest date supported. """
        return cls.DATE_TYPE.max

    def min_timestamp(self, **kwargs):
        """
        First timestamp of the range, or None if empty.
        :return: a datetime.datetime object
        """
        if self:
            return self.start.min_timestamp(**kwargs)

    def max_timestamp(self, **kwargs):
        """
        Last timestamp of the range, or None if empty.
        :return: a datetime.datetime object
        """
        if self:
            # note: self.stop-1 is not necessarily in self, if step>1. thus using self[-1]
            return self[-1].max_timestamp(**kwargs)

    def cli_repr(self):
        """
        A command-line representation of the date range.
        This is a convenience method for providing the date range to another script.
        """
        if len(self) == 1:
            return str(self[0])
        return self.to_string('', '')

    # ===========================================================================
    # Intersection
    # ===========================================================================

    def __and__(self, other):
        return self.intersection(other)

    def intersection(self, other):
        """
        Calculate the intersection of the two DateRanges and return the result as a DateRange.

        This implementation was copied (and adjusted) from rangeplus
        (https://github.com/avnr/rangeplus).
        """

        # check steps have the same sign
        if (self.step > 0) != (other.step > 0):
            raise ValueError('Intersection is undefined for steps with opposite signs: %r & %r' % (
                self, other))

        # more helpers

        def stop_min(x, y):
            return None if x is None and y is None \
                else x if y is None else y if x is None else min(x, y)

        def stop_max_inv(x, y):
            return None if x is None and y is None \
                else x if y is None else y if x is None else max(x, y)

        # return empty Range if either ranges is empty
        empty = self.empty()
        if not self or not other:
            return empty()

        # both directions are the same
        step0, step1, sign, offset = (abs(self.step), abs(other.step),
                                      (self.step > 0) - (self.step < 0), other.start - self.start)
        gcd, x, y = _egcd(step0, step1)
        interval0, interval1 = step0 // gcd, step1 // gcd  # calculate the coprime intervals
        step = interval0 * interval1 * gcd * sign
        if offset % gcd != 0:  # return empty result if offset not alligned on gcd
            return empty()

        # Apply Chinese Remainder Theorem
        # x % interval1 means inverse_mod(interval0, interval1)
        crt = (offset * interval0 * (x % interval1)) % step
        filler = 0
        if sign > 0 and offset > 0 or sign < 0 and offset < 0:
            gap = offset - crt
            filler = gap if 0 == gap % step else (gap // step + 1) * step
        start = self.start + crt + filler
        stop = stop_min(self.stop, other.stop) if sign > 0 else stop_max_inv(self.stop, other.stop)
        return type(self)(start, stop, step)

    # ===========================================================================
    # privates
    # ===========================================================================

    def _o2d(self, ordinal):
        return self.DATE_TYPE.fromordinal(ordinal)

    def _d2o(self, date):
        # Note: supports Date, datetime.date, and datetime.datetime
        return date.toordinal()


################################################################################



"""
Convenient Date and DateRange classes.
"""

from .date import Date
from .range import DateRange

# we import .argparse, for the side effect
from . import argparse as _argparse

Date, DateRange, _argparse  # pyflakes



"""
Misc tools used in this pacakge.
"""

import functools


class classproperty(object):

    def __init__(self, fget):
        self.fget = classmethod(fget)
        functools.update_wrapper(self, fget)

    def __get__(self, *a):
        return self.fget.__get__(*a)()


def egcd(a, b):
    """ Return a triple (g, x, y), where ax + by = g = gcd(a, b) """
    x, y, u, v = 0, 1, 1, 0
    while a:
        q, r = b // a, b % a
        m, n = x - u * q, y - v * q
        b, a, x, y, u, v = a, r, u, v, m, n
    return b, x, y



"""
Timezone stuff.
"""

import datetime
import pytz


TIMEZONE = None
"""
Timezone used for resolving what "today" means.  Defaults to local tz.
"""


def get_timezone():
    return TIMEZONE


def set_timezone(tz):
    """
    Set a timezone to use for resolving what "today" means.
    Set to None to reset to default.
    """
    global TIMEZONE
    if tz is not None and not isinstance(tz, datetime.tzinfo):
        tz = pytz.timezone(tz)
    prev_tz = TIMEZONE
    TIMEZONE = tz
    return prev_tz


def set_timezone_utc():
    """
    Same as ``set_timezone('UTC')``.
    """
    return set_timezone('UTC')



#!/usr/bin/env python3

from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

# Read version info from yyyymmdd/version.py
version_vars = {}
with open("yyyymmdd/version.py") as fp:
    exec(fp.read(), version_vars)
version_string = version_vars['__version_string__']

setup(
    name='yyyymmdd',
    version=version_string,

    description='Convenient Date and DateRange classes.',
    long_description=long_description,
    url='https://github.com/shx2/yyyymmdd',
    author='shx2',
    author_email='shx222@gmail.com',
    license='MIT',

    packages=find_packages(exclude=['tests*']),
    platforms = ["POSIX", "Windows"],
    install_requires=[],

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Programming Language :: Python :: 3',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development',
        'License :: OSI Approved :: MIT License',
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
    ],
    keywords='date datetime range',
)
